from django.conf import settings
from django.core.exceptions import FieldError
from django.db import models
from django.db.models import DO_NOTHING
from six import python_2_unicode_compatible

from .app_settings import PWA_APP_NAME


@python_2_unicode_compatible
class Group(models.Model):
    name = models.CharField(max_length=255, unique=True)

    def __str__(self):
        return "Push Group: %s" % self.name


@python_2_unicode_compatible
class SubscriptionInfo(models.Model):
    browser = models.CharField(max_length=100)
    endpoint = models.URLField(max_length=255)
    auth = models.CharField(max_length=100)
    p256dh = models.CharField(max_length=100)

    def __str__(self):
        return "Subscription Information for %s" % self.browser


@python_2_unicode_compatible
class PushInformation(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="webpush_info", blank=True, null=True, on_delete=models.CASCADE
    )
    subscription = models.ForeignKey(SubscriptionInfo, related_name="webpush_info", on_delete=models.CASCADE)
    group = models.ForeignKey(Group, related_name="webpush_info", blank=True, null=True, on_delete=models.CASCADE)

    def save(self, *args, **kwargs):
        # Check whether user or the group field is present
        # At least one field should be present there
        # Through from the functionality its not possible, just in case! ;)
        if self.user or self.group:
            super(PushInformation, self).save(*args, **kwargs)
        else:
            raise FieldError("At least user or group should be present")

    def __str__(self):

        return "Push Subscription for: %s" % self.user


@python_2_unicode_compatible
class PushMessage(models.Model):
    active = models.BooleanField(default=True)
    send_on = models.DateTimeField(auto_now_add=True, null=True)
    title = models.CharField(max_length=200, null=False, default="Message from %s." % PWA_APP_NAME)
    message = models.TextField(blank=True, null=True)
    url = models.URLField(blank=True, null=True)
    icon = models.URLField(blank=True, null=True)
    send_to = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.CASCADE)
    sent = models.BooleanField(default=False)

    def __str__(self):
        if self.sent:
            retn = "%s, sent on %s" % (self.title, self.send_on)

        else:
            retn = "%s, to be sent on %s" % (self.title, self.send_on)

        return retn



import json

from django.contrib import admin
from django.utils import timezone

from pwa_webpush import send_user_notification
from .custom_site_class import CustomSite
from .models import PushInformation, PushMessage
from .utils import _send_notification


@admin.register(PushInformation)
class PushInfoAdmin(admin.ModelAdmin):
    list_display = ("__str__", "user", "subscription", "group")
    actions = ("send_test_message",)

    def send_test_message(self, request, queryset):
        result = []
        payload = {"head": "Hey", "body": "Hello World"}
        for device in queryset:
            result.append(_send_notification(device, json.dumps(payload), 0))


@admin.register(PushMessage)
class PwaPushMessageAdmin(admin.ModelAdmin):
    def save_model(self, request, obj, form, change):
        site = CustomSite(request)
        obj.save()

        if not obj.sent and obj.active and obj.send_on <= timezone.now():
            payload = {
                "head": obj.title,
                "body": obj.message,
                "icon": site.get_external_url("/static/pwa/icons/apple-icon-72x72.png") if not obj.icon else obj.icon,
                "url": site.get_external_url("/m/") if not obj.url else obj.url,
            }

            if not obj.send_to:
                push_users = PushInformation.objects.all()

            else:
                push_users = PushInformation.objects.filter(user=obj.send_to)

            for push_user in push_users:
                send_user_notification(user=push_user.user, payload=payload, ttl=1000)
                obj.sent = True
                obj.save()

    list_display = (
        u"active",
        u"send_on",
        u"title",
        u"message",
        u"url",
        u"icon",
    )
    list_filter = (u"active", u"send_on")



from django.conf import settings
from django.forms.models import model_to_dict
from django.urls import reverse

from pywebpush import WebPushException, webpush


def send_notification_to_user(user, payload, ttl=0):
    # Get all the push_info of the user

    errors = []
    push_infos = user.webpush_info.select_related("subscription")
    for push_info in push_infos:
        try:
            _send_notification(push_info.subscription, payload, ttl)

        except WebPushException as ex:
            errors.append(dict(subscription=push_info.subscription, exception=ex))

    if errors:
        raise WebPushException("Push failed.")


def send_notification_to_group(group_name, payload, ttl=0):
    from .models import Group

    # Get all the subscription related to the group

    errors = []
    push_infos = Group.objects.get(name=group_name).webpush_info.select_related("subscription")
    for push_info in push_infos:
        try:
            _send_notification(push_info.subscription, payload, ttl)

        except WebPushException as ex:
            errors.append(dict(subscription=push_info.subscription, exception=ex))

    if errors:
        raise WebPushException("Push failed.")


def send_to_subscription(subscription, payload, ttl=0):
    _send_notification(subscription, payload, ttl)


def _send_notification(subscription, payload, ttl):
    subscription_data = _process_subscription_info(subscription)
    vapid_data = {}

    webpush_settings = getattr(settings, "WEBPUSH_SETTINGS", {})
    vapid_private_key = webpush_settings.get("VAPID_PRIVATE_KEY")
    vapid_admin_email = webpush_settings.get("VAPID_ADMIN_EMAIL")

    # Vapid keys are optional, and mandatory only for Chrome.
    # If Vapid key is provided, include vapid key and claims
    if vapid_private_key:
        vapid_data = {
            "vapid_private_key": vapid_private_key,
            "vapid_claims": {"sub": "mailto:{}".format(vapid_admin_email)},
        }

    req = webpush(subscription_info=subscription_data, data=payload, ttl=ttl, **vapid_data)
    return req


def _process_subscription_info(subscription):
    subscription_data = model_to_dict(subscription, exclude=["browser", "id"])
    endpoint = subscription_data.pop("endpoint")
    p256dh = subscription_data.pop("p256dh")
    auth = subscription_data.pop("auth")

    return {"endpoint": endpoint, "keys": {"p256dh": p256dh, "auth": auth}}


def get_templatetag_context(context):
    request = context["request"]
    vapid_public_key = getattr(settings, "WEBPUSH_SETTINGS", {}).get("VAPID_PUBLIC_KEY", "")

    data = {
        "group": context.get("webpush", {}).get("group"),
        "user": getattr(request, "user", None),
        "vapid_public_key": vapid_public_key,
        "webpush_save_url": reverse("save_webpush_info"),
    }

    return data



import json

from .utils import send_notification_to_group, send_notification_to_user

default_app_config = "pwa_webpush.apps.PwaConfig"


def send_group_notification(group_name, payload, ttl=0):
    payload = json.dumps(payload)
    send_notification_to_group(group_name, payload, ttl)


def send_user_notification(user, payload, ttl=0):
    payload = json.dumps(payload)
    send_notification_to_user(user, payload, ttl)



from django.urls import path

from .views import manifest, offline, ServiceWorkerView, save_info

# Serve up serviceworker.js and manifest.json at the root
urlpatterns = [
    path("serviceworker.js", ServiceWorkerView.as_view(), name="serviceworker"),
    path("manifest.json", manifest, name="manifest"),
    path("offline/", offline, name="offline"),
    path("webpush/save_information", save_info, name="save_webpush_info"),
]



# Generated by Django 2.2.1 on 2019-05-08 19:08

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Group',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='SubscriptionInfo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('browser', models.CharField(max_length=100)),
                ('endpoint', models.URLField(max_length=255)),
                ('auth', models.CharField(max_length=100)),
                ('p256dh', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='PushMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('active', models.BooleanField(default=True)),
                ('send_on', models.DateTimeField(auto_now_add=True, null=True)),
                ('title', models.CharField(default='Message from MyApp.', max_length=200)),
                ('message', models.TextField(blank=True, null=True)),
                ('url', models.URLField(blank=True, null=True)),
                ('icon', models.URLField(blank=True, null=True)),
                ('sent', models.BooleanField(default=False)),
                ('send_to', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='PushInformation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='webpush_info', to='pwa_webpush.Group')),
                ('subscription', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='webpush_info', to='pwa_webpush.SubscriptionInfo')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='webpush_info', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



""" Settings required by django-app. """
import os

from django.conf import settings

# Path to the service worker implementation.  Default implementation is empty.
PWA_SERVICE_WORKER_PATH = getattr(
    settings,
    "PWA_SERVICE_WORKER_PATH",
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "templates", "serviceworker.js"),
)
# App parameters to include in manifest.json and appropriate meta tags
PWA_APP_NAME = getattr(settings, "PWA_APP_NAME", "MyApp")
PWA_APP_DESCRIPTION = getattr(settings, "PWA_APP_DESCRIPTION", "My Progressive Web App")
PWA_APP_ROOT_URL = getattr(settings, "PWA_APP_ROOT_URL", "/")
PWA_APP_THEME_COLOR = getattr(settings, "PWA_APP_THEME_COLOR", "#000")
PWA_APP_BACKGROUND_COLOR = getattr(settings, "PWA_APP_BACKGROUND_COLOR", "#fff")
PWA_APP_DISPLAY = getattr(settings, "PWA_APP_DISPLAY", "standalone")
PWA_APP_SCOPE = getattr(settings, "PWA_APP_SCOPE", "/")
PWA_APP_ORIENTATION = getattr(settings, "PWA_APP_ORIENTATION", "any")
PWA_APP_START_URL = getattr(settings, "PWA_APP_START_URL", "/")
PWA_APP_FETCH_URL = getattr(settings, "PWA_APP_FETCH_URL", "/")
PWA_APP_ICONS = getattr(
    settings,
    "PWA_APP_ICONS",
    [
        {"src": "/static/images/icons/icon-72x72.png", "size": "72x72"},
        {"src": "/static/images/icons/icon-96x96.png", "size": "96x96"},
        {"src": "/static/images/icons/icon-128x128.png", "size": "128x128"},
        {"src": "/static/images/icons/icon-144x144.png", "size": "144x144"},
        {"src": "/static/images/icons/icon-152x152.png", "size": "152x152"},
        {"src": "/static/images/icons/icon-192x192.png", "size": "192x192"},
        {"src": "/static/images/icons/icon-384x384.png", "size": "384x384"},
        {"src": "/static/images/icons/icon-512x512.png", "size": "512x512"},
    ],
)
PWA_APP_SPLASH_SCREEN = getattr(
    settings,
    "PWA_APP_SPLASH_SCREEN",
    [
        {
            "src": "/static/images/icons/splash-640x1136.png",
            "media": "(device-width: 320px) and (device-height: 568px) and (-webkit-device-pixel-ratio: 2)",
        },
        {
            "src": "/static/images/icons/splash-750x1334.png",
            "media": "(device-width: 375px) and (device-height: 667px) and (-webkit-device-pixel-ratio: 2)",
        },
        {
            "src": "/static/images/icons/splash-1242x2208.png",
            "media": "(device-width: 621px) and (device-height: 1104px) and (-webkit-device-pixel-ratio: 3)",
        },
        {
            "src": "/static/images/icons/splash-1125x2436.png",
            "media": "(device-width: 375px) and (device-height: 812px) and (-webkit-device-pixel-ratio: 3)",
        },
        {
            "src": "/static/images/icons/splash-828x1792.png",
            "media": "(device-width: 414px) and (device-height: 896px) and (-webkit-device-pixel-ratio: 2)",
        },
        {
            "src": "/static/images/icons/splash-1242x2688.png",
            "media": "(device-width: 414px) and (device-height: 896px) and (-webkit-device-pixel-ratio: 3)",
        },
        {
            "src": "/static/images/icons/splash-1536x2048.png",
            "media": "(device-width: 768px) and (device-height: 1024px) and (-webkit-device-pixel-ratio: 2)",
        },
        {
            "src": "/static/images/icons/splash-1668x2224.png",
            "media": "(device-width: 834px) and (device-height: 1112px) and (-webkit-device-pixel-ratio: 2)",
        },
        {
            "src": "/static/images/icons/splash-1668x2388.png",
            "media": "(device-width: 834px) and (device-height: 1194px) and (-webkit-device-pixel-ratio: 2)",
        },
        {
            "src": "/static/images/icons/splash-2048x2732.png",
            "media": "(device-width: 1024px) and (device-height: 1366px) and (-webkit-device-pixel-ratio: 2)",
        },
    ],
)
PWA_APP_DIR = getattr(settings, "PWA_APP_DIR", "auto")
PWA_APP_LANG = getattr(settings, "PWA_APP_LANG", "en-US")



from urllib.parse import urlsplit, parse_qs

from django.urls import reverse


class CustomSite(object):
    protocol = "http"
    domain = None
    request_path = None
    fragment = None
    query = None
    params = None
    port = None

    site_url = None  # Without Slash
    base_url = None  # With Slash
    canonical_url = None

    def __init__(self, request):
        request_uri = request.build_absolute_uri()

        url = urlsplit(request_uri)

        self.protocol = url.scheme
        self.domain = url.hostname
        self.port = url.port
        self.request_path = url.path
        self.fragment = url.fragment
        self.query = url.query
        self.params = parse_qs(self.query)

        # Without Slash
        if self.port and self.port not in [80, 443]:
            self.site_url = "%s://%s:%s" % (self.protocol, self.domain, str(self.port))
        else:
            self.site_url = "%s://%s" % (self.protocol, self.domain)

        self.base_url = "%s/" % self.site_url  # With slash
        self.canonical_url = self.get_external_url(self.request_path)

    def get_external_url(self, resource_path):
        """
        Returns a fully qualified web address fo a resource.

        :param resource_path: A site relative path.
        :return: A url that includes the full site url to the resource
        """

        if resource_path[0] in ["/", "\\"]:
            urlbase = self.site_url
        else:
            urlbase = self.base_url

        external_url = "%s%s" % (urlbase, resource_path)

        return external_url

    def external_reverse(self, url_slug, **kwargs):

        return self.get_external_url(reverse(url_slug))



from django import forms

from .models import Group, PushInformation, SubscriptionInfo


class WebPushForm(forms.Form):
    group = forms.CharField(max_length=255, required=False)
    status_type = forms.ChoiceField(choices=[("subscribe", "subscribe"), ("unsubscribe", "unsubscribe")])

    def save_or_delete(self, subscription, user, status_type, group_name):
        data = {}

        if user.is_authenticated:
            data["user"] = user

        if group_name:
            group, created = Group.objects.get_or_create(name=group_name)
            data["group"] = group

        data["subscription"] = subscription

        push_info, created = PushInformation.objects.get_or_create(**data)

        # If unsubscribe is called, that means need to delete the browser
        # and notification info from server.
        if status_type == "unsubscribe":
            push_info.delete()
            subscription.delete()


class SubscriptionForm(forms.ModelForm):
    class Meta:
        model = SubscriptionInfo
        fields = ("endpoint", "auth", "p256dh", "browser")

    def get_or_save(self):
        subscription, created = SubscriptionInfo.objects.get_or_create(**self.cleaned_data)
        return subscription



import json

from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from django.views.generic import TemplateView

from pwa_webpush.forms import SubscriptionForm, WebPushForm
from . import app_settings
from .custom_site_class import CustomSite


def manifest(request):
    context = {
        setting_name: getattr(app_settings, setting_name)
        for setting_name in dir(app_settings)
        if setting_name.startswith("PWA_")
    }

    site = CustomSite(request)

    if request.scheme not in context.get("PWA_APP_START_URL"):
        context["PWA_APP_START_URL"] = site.get_external_url(context.get("PWA_APP_START_URL"))

    if request.scheme not in context.get("PWA_APP_SCOPE"):
        context["PWA_APP_SCOPE"] = site.get_external_url(context.get("PWA_APP_SCOPE"))

    return render(request, "manifest.json", context)


def offline(request):
    return render(request, "offline.html")


def process_subscription_data(post_data):
    """Process the subscription data according to out model"""
    subscription_data = post_data.pop("subscription", {})
    # As our database saves the auth and p256dh key in separate field,
    # we need to refactor it and insert the auth and p256dh keys in the same dictionary
    keys = subscription_data.pop("keys", {})
    subscription_data.update(keys)
    # Insert the browser name
    subscription_data["browser"] = post_data.pop("browser")
    return subscription_data


@require_POST
@csrf_exempt
def save_info(request):
    # Parse the  json object from post data. return 400 if the json encoding is wrong
    try:
        post_data = json.loads(request.body.decode("utf-8"))
    except ValueError:
        return HttpResponse(status=400)

    # Process the subscription data to mach with the model
    subscription_data = process_subscription_data(post_data)
    subscription_form = SubscriptionForm(subscription_data)
    # pass the data through WebPushForm for validation purpose
    web_push_form = WebPushForm(post_data)

    # Check if subscriptioninfo and the web push info bot are valid
    if subscription_form.is_valid() and web_push_form.is_valid():
        # Get the cleaned data in order to get status_type and group_name
        web_push_data = web_push_form.cleaned_data
        status_type = web_push_data.pop("status_type")
        group_name = web_push_data.pop("group")

        # We at least need the user or group to subscribe for a notification
        if request.user.is_authenticated or group_name:
            # Save the subscription info with subscription data
            # as the subscription data is a dictionary and its valid
            subscription = subscription_form.get_or_save()
            web_push_form.save_or_delete(
                subscription=subscription, user=request.user, status_type=status_type, group_name=group_name
            )

            # If subscribe is made, means object is created. So return 201
            if status_type == "subscribe":
                return HttpResponse(status=201)
            # Unsubscribe is made, means object is deleted. So return 202
            elif "unsubscribe":
                return HttpResponse(status=202)

    return HttpResponse(status=400)


class ServiceWorkerView(TemplateView):
    """
    Service Worker need to be loaded from same domain.
    Therefore, use TemplateView in order to server the webpush_serviceworker.js
    """

    template_name = "serviceworker.js"
    content_type = "application/javascript"



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

install_requirements = [
    "django>=1.8",
    "six",
]

setup(
    name='django-pwa-webpush',
    version='1.1.2',
    packages=find_packages(),
    include_package_data=True,
    license='GNU Public License',
    description='A conglomeration of django-pwa and django-webpush.',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://www.github.com/avryhof/django-pwa-webpush',
    author='Amos Vryhof',
    author_email='amos@vryhofresearch.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'pywebpush==1.6.0'
    ]
)
